from datetime import datetime, date, timedelta, time
from decimal import Decimal
from json import JSONEncoder, dumps, loads
from base64 import b64encode, b64decode
from pydicts import casts

# Forma en que debe parsearse los Decimals
class DecimalsWay:
    Decimal=1  #Uses a String with decimal to detect decimals
    String=2
    Float=3

## Usa
class MyJSONEncoder(JSONEncoder):
    # Part of this code is from https://github.com/django/django/blob/main/django/core/serializers/json.py
    # JSONEncoder subclass that knows how to encode date/time, decimal types, and UUIDs.
    def _get_duration_components(self, duration):
        days = duration.days
        seconds = duration.seconds
        microseconds = duration.microseconds

        minutes = seconds // 60
        seconds %= 60


        hours = minutes // 60
        minutes %= 60

        return days, hours, minutes, seconds, microseconds
    
    def _duration_iso_string(self, duration):
        if duration < timedelta(0):
            sign = "-"
            duration *= -1
        else:
            sign = ""

        days, hours, minutes, seconds, microseconds = self._get_duration_components(duration)
        ms = ".{:06d}".format(microseconds) if microseconds else ""
        return "{}P{}DT{:02d}H{:02d}M{:02d}{}S".format(
            sign, days, hours, minutes, seconds, ms
        )
    
    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime):
            return o.isoformat().replace("+00:00","Z")
        elif isinstance(o, date):
            return o.isoformat()
        elif isinstance(o, time):
            if o.utcoffset() is not None: #If it's aware
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r
        elif isinstance(o, timedelta):
            return self._duration_iso_string(o)
        elif isinstance(o, Decimal):
            return f"Decimal('{o}')"
        elif o.__class__.__name__ in ("Promise", "__proxy__"): #django.utils.functional
            return str(o)
        elif isinstance(o, bytes):
            return b64encode(o).decode("UTF-8")
        elif o.__class__.__name__=="Percentage":
            return o.value
        elif o.__class__.__name__=="Currency":
            return o.amount
        else:
            return super().default(o)

## Usa decimals como floats normalmwente para JS
class MyJSONEncoderDecimalsAsString(MyJSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return str(o)
        else:
            return super().default(o)
            
## Usa decimals como floats normalmwente para JS
class MyJSONEncoderDecimalsAsFloat(MyJSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return float(o)
        else:
            return super().default(o)

                
def MyJSONEncoder_dumps(r, indent=4):
    return dumps(r, cls=MyJSONEncoder, indent=indent)
                
def MyJSONEncoderDecimalsAsString_dumps(r, indent=4):
    return dumps(r, cls=MyJSONEncoderDecimalsAsString, indent=indent)
                
def MyJSONEncoderDecimalsAsFloat_dumps(r, indent=4):
    return dumps(r, cls=MyJSONEncoderDecimalsAsFloat, indent=indent)


def MyJSONEncoder_loads(s):
    def hooks_MyJSONEncoder(iter_value):
        return hooks(iter_value, DecimalsWay.Decimal)
    ##############################
    return loads(s, object_hook=hooks_MyJSONEncoder)

def MyJSONEncoderDecimalsAsFloat_loads(s):
    
    def hooks_MyJSONEncoderAsFloat(iter_value):
        return hooks(iter_value, DecimalsWay.Float)
    ####################################
    return loads(s, object_hook=hooks_MyJSONEncoderAsFloat)        

def MyJSONEncoderDecimalsAsString_loads(s):
    def hooks_MyJSONEncoderAsString(iter_value):
        return hooks(iter_value, DecimalsWay.String)
    ######################################
    return loads(s, object_hook=hooks_MyJSONEncoderAsString)
    
    
def hooks(iter_value, decimals_way):
    """
        Iterates a dict or list to cast decimals and dtaware in json.loads using objeck_hook
    """
    
    def get_date(s):
        try:
            return date.fromisoformat(s)
        except:
            return None 
            
    def get_dtaware(s):
        try:
            print(s, casts.str2dtaware(s,"JsUtcIso"))
            return casts.str2dtaware(s,"JsUtcIso")
        except:
            return None 

    def get_dtnaive(s):
        try:
            print(s, casts.str2dtnaive(s,"JsIso"))
            return casts.str2dtnaive(s,"JsIso")
        except:
            return None 
            
    #    Parse the ISO8601 duration string as hours, minutes, seconds
    def get_timedelta(str):
    #    try:
    ## https://stackoverflow.com/questions/36976138/is-there-an-easy-way-to-convert-iso-8601-duration-to-timedelta
    ## Parse the ISO8601 duration as years,months,weeks,days, hours,minutes,seconds
    ## Returns: milliseconds
    ## Examples: "PT1H30M15.460S", "P5DT4M", "P2WT3H"
        def get_isosplit(str, split):
            if split in str:
                n, str = str.split(split, 1)
            else:
                n = '0'
            return n.replace(',', '.'), str  # to handle like "P0,5Y"

        str = str.split('P', 1)[-1]  # Remove prefix
        s_yr, str = get_isosplit(str, 'Y')  # Step through letter dividers
        s_mo, str = get_isosplit(str, 'M')
        s_wk, str = get_isosplit(str, 'W')
        s_dy, str = get_isosplit(str, 'D')
        _, str    = get_isosplit(str, 'T')
        s_hr, str = get_isosplit(str, 'H')
        s_mi, str = get_isosplit(str, 'M')
        s_sc, str = get_isosplit(str, 'S')
        n_yr = float(s_yr) * 365   # approx days for year, month, week
        n_mo = float(s_mo) * 30.4
        n_wk = float(s_wk) * 7
        dt = datetime.timedelta(days=n_yr+n_mo+n_wk+float(s_dy), hours=float(s_hr), minutes=float(s_mi), seconds=float(s_sc))
        print(dt)
        return int(dt.total_seconds()*1000) ## int(dt.total_seconds()) | dt
    #    except:
    #        return None 
            
    def get_time(s):
        try:
            if not ":" in s:
                return None
            return time.fromisoformat(s)
        except:
            return None 
            
    def get_bytes(s):
        try:
            return b64decode(s)
        except:
            return None
            
    def get_Decimal(s):
        try:
            return eval(s)
        except:
            return None

        
    def guess_cast(o, decimal_way):        
        if decimal_way==DecimalsWay.Decimal:
            r=get_Decimal(o)
            if r is not None:
                return  r

        r=get_date(o)
        if r is not None:
            return  r
            
        r=get_dtnaive(o)
        if r is not None:
            return  r
            
        r=get_dtaware(o)
        if r is not None:
            return  r
            
        r=get_time(o)
        if r is not None:
            return  r
            
        r=get_bytes(o)
        if r is not None:
            return  r
        
        return o
    ########################################################
    if isinstance(iter_value, dict):
        for k, v in iter_value.items():
            if isinstance(v, dict):
                iter_value[k]=hooks(v, decimals_way)
            elif isinstance(iter_value, list):
                for i in v:
                    i=hooks(v, decimals_way)
            else:
                iter_value[k]=guess_cast(v, decimals_way)
    elif isinstance(iter_value, list):
        for i in v:
            i=hooks(i, decimals_way)
    return iter_value
    
    




### If a function only can be used by dtaware or naive it will have its prefix dtaware_ or dtnaive_
### If a function can use both of them its prefix will be dt_
from decimal import Decimal
from datetime import timedelta, date, datetime, time
from gettext import translation
from importlib.resources import files
from pydicts import exceptions
from zoneinfo import ZoneInfo
        
try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str


def object_or_empty(v):
    """
        Returns and empty string if None, else return value
    """
    return "" if v is None else v

## Converts a string  to a decimal
def str2decimal(s, type):
    if s is None:
        return None
    if type==1: #2.123,25
        try:
            return Decimal(s.replace(".","").replace(",", "."))
        except:
            raise exceptions.CastException(_("Method str2decimal couln't convert {0} ({1}) to a Decimal"))


def str2bool(value):
    """
        Converts strings True or False to boolean
        @param s String
        @return Boolean
    """
    def exception():
        raise exceptions.CastException(_("Method str2bool couldn't convert {0} ({1}) to a boolean").format(value, value.__class__))
    if not value.__class__ is str:
        exception()
        
    if value=="0" or value.lower()=="false":
        return False
    elif value=="1" or value.lower()=="true":
        return True
    exception()
### Function that converts a None value into a Decimal('0')
### @param dec Should be a Decimal value or None
### @return Decimal
#def none2decimal0(dec):
#    return none2alt(dec,Decimal('0'))
#
### If a value is None, returns an alternative
#def none2alt(value, alternative):
#    if value==None:
#        return alternative
#    return value
#
def bytes2str(b, code='UTF-8'):
    """
        Bytes 2 string
    """ 
    if b is None:
        return None
    return b.decode(code)
    
def str2bytes(s, code='UTF8'):
    """
        String 2 bytes
    """
    if s is None:
        return None
    else:
        return s.encode(code)

### Returns if a datetime is aware
def is_aware(dt):
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        return False
    return True

## Returns if a datetime is naive
def is_naive(dt):
    return not is_aware(dt)

## Function to create a datetime aware object
## @param date datetime.date object
## @param hour hour object
## @param tz_name String with datetime tz_name name. For example "Europe/Madrid"
## @return datetime aware
def dtaware(date_, time_, tz_name):
    dt_naive=dtnaive(date_, time_)
    return dtnaive2dtaware(dt_naive, tz_name)

def dtnaive2dtaware(dtnaive, tz_name):
    return dtnaive.replace(tzinfo=ZoneInfo(tz_name))


def dtaware_now(tzname=None):
    """
        If tzname is None: returns UTC dtaware
    """
    utc_aware=dtnaive2dtaware(dtnaive_now(), 'UTC')
    if tzname is None:
        return utc_aware
    else:
        return dtaware_changes_tz(utc_aware, tzname)

def dtnaive_now():
    return datetime.now()

## Function to create a datetime aware object
## @param date datetime.date object
## @param hour hour object
## @return datetime naive
def dtnaive(date_, hour):
    return datetime(date_.year,  date_.month,  date_.day,  hour.hour,  hour.minute,  hour.second, hour.microsecond)


## Returns a date with the first date of the month
## @param year Year to search fist day
## @param month Month to search first day
def date_first_of_the_month(year, month):
    return date(year, month, 1)

## Returns a date with the last date of the month
## @param year Year to search last day
## @param month Month to search last day
def date_last_of_the_month(year, month):
    if month==12:
        return date(year, month, 31)
    return date(year, month+1, 1)-timedelta(days=1)

## Returns a date with the first date of the year
## @param year Year to search first day
def date_first_of_the_year(year):
    return date_first_of_the_month(year, 1)

## Returns a date with the last date of the year
## @param year Year to search last day
def date_last_of_the_year(year):
    return date_last_of_the_month(year,12)

## Returns a date with the first date of the month after x months
## @param year Year to search  day
## @param month Month to search day
## @param x Number of months after parameters. Cab be positive to add months or negative to substract months
def date_first_of_the_next_x_months(year, month, x):
    if x>=0:
        first=date(year, month, 1)
        for i in range(x):
            last=date_last_of_the_month(first.year, first.month)
            first=last+timedelta(days=1)
        return first
    else:#<0
        first=date(year, month, 1)
        for i in range(abs(x)):
            last_before=first-timedelta(days=1)
            first=date_first_of_the_month(last_before.year, last_before.month)
        return first    

## Returns a date with the last date of the month after x months
## @param year Year to search  day
## @param month Month to search day
## @param x Number of months after parameters. Cab be positive to add months or negative to substract months
def date_last_of_the_next_x_months(year, month, x):
    first=date_first_of_the_next_x_months(year, month, x)
    return date_last_of_the_month(first.year, first.month)

def dtaware_month_end(year, month, tz_name):
    return dtaware_day_end_from_date(date_last_of_the_month(year, month), tz_name)
    
## Returns an aware datetime with the start of year
def dtaware_year_start(year, tz_name):
    return dtaware_day_start_from_date(date(year, 1, 1), tz_name)
    
## Returns an aware datetime with the last of year
def dtaware_year_end(year, tz_name):
    return dtaware_day_end_from_date(date(year, 12, 31), tz_name)
    
def dtaware_day_end(dt, tz_name):
    """
        Returns the last  datetime (microsecond  level) of the  day in tz_name zone
    """
    if is_naive():
        raise exceptions.CastException(_("A datetime with timezone is needed"))
    dt=dtaware_changes_tz(dt, tz_name)
    return dt.replace(hour=23, minute=59, second=59, microsecond=999999)
    
def dtnaive_day_end(dt):
    """
        Returns the last  datetime (microsecond  level) of the  day in naive format
    """
    if is_aware(dt):
        raise exceptions.CastException(_("A datetime without timezone is needed"))
    return dt.replace(hour=23, minute=59, second=59, microsecond=999999)

## Returns the end of the day dtnaive from a date
def dtnaive_day_end_from_date(date_):
    dt=datetime(date_.year, date_.month, date_.day)
    return dtnaive_day_end(dt)

## Returns the end of the day dtaware in utc format
def dtaware_day_end_from_date(date, tz_name):
    return dtaware(date, time(23, 59, 59, 999999), tz_name)
    
## Returns a dtnaive or dtawre (as parameter) with the end of the day
def dtnaive_day_start(dt):
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)

## Returns a dtnaive or dtawre (as parameter) with the end of the day in zone tz_name
def dtaware_day_start(dt, tz_name):
    if is_naive():
        raise exceptions.CastException(_("A datetime with timezone is needed"))
    dt=dtaware_changes_tz(dt, tz_name)
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)

## Returns the end of the day dtnaive from a date
def dtnaive_day_start_from_date(date_):
    dt=datetime(date_.year, date_.month, date_.day)
    return dtnaive_day_start(dt)

## Returns the end of the day dtaware of the tz_name timezone from a date
def dtaware_day_start_from_date(date, tz_name):
    return dtaware(date, time(0, 0, 0, 0), tz_name)

## Returns the start of a month in utc format
def dtaware_month_start(year, month, tz_name):
    return dtaware_day_start_from_date(date(year, month, 1), tz_name)
#    
#def month2int(s):
#    """
#        Converts a month string to a int
#    """
#    if s in ["Jan", "Ene", "Enero", "January", "enero", "january"]:
#        return 1
#    if s in ["Feb", "Febrero", "February", "febrero", "february"]:
#        return 2
#    if s in ["Mar", "Marzo", "March", "marzo", "march"]:
#        return 3
#    if s in ["Apr", "Abr", "April", "Abril", "abril", "april"]:
#        return 4
#    if s in ["May", "Mayo", "mayo", "may"]:
#        return 5
#    if s in ["Jun", "June", "Junio", "junio", "june"]:
#        return 6
#    if s in ["Jul", "July", "Julio", "julio", "july"]:
#        return 7
#    if s in ["Aug", "Ago", "August", "Agosto", "agosto", "august"]:
#        return 8
#    if s in ["Sep", "Septiembre", "September", "septiembre", "september"]:
#        return 9
#    if s in ["Oct", "October", "Octubre", "octubre", "october"]:
#        return 10
#    if s in ["Nov", "Noviembre", "November", "noviembre", "november"]:
#        return 11
#    if s in ["Dic", "Dec", "Diciembre", "December", "diciembre", "december"]:
#        return 12

def str2time(s, format="HH:MM"):
    allowed=["HH:MM", "HH:MM:SS","HH:MMxx"]
    if format in allowed:
        if format=="HH:MM":#12:12
            a=s.split(":")
            return time(int(a[0]), int(a[1]))
        elif format=="HH:MM:SS":#12:12:12
            a=s.split(":")
            return time(int(a[0]), int(a[1]), int(a[2]))
        elif format=="HH:MMxx": #5:12am o pm
            s=s.upper()
            s=s.replace("AM", "")
            if s.find("PM"):
                s=s.replace("PM", "")
                points=s.split(":")
                return time(int(points[0])+12, int(points[1]))
            else:#AM
                points=s.split(":")
                return time(int(points[0]), int(points[1]))
    else:
        raise exceptions.CastException(_("I can't convert this format '{}'. I only support this {}").format(format, allowed))

## Converts a time to a string
def time2str(ti, format="HH:MM" ):
    allowed=["HH:MM", "HH:MM:SS","Xulpymoney"]
    if format in allowed:
        if ti==None:
            return None
        if format=="Xulpymoney":
            if ti.microsecond in (4, 5):
                return str(ti)[11:-13]
            else:
                return str(ti)[11:-6]
        elif format=="HH:MM":
            return ("{}:{}".format(str(ti.hour).zfill(2), str(ti.minute).zfill(2)))
        elif format=="HH:MM:SS":
            return ("{}:{}:{}".format(str(ti.hour).zfill(2), str(ti.minute).zfill(2), str(ti.second).zfill(2)))


def str2date(iso, format="YYYY-MM-DD"):
    allowed=["YYYY-MM-DD", "DD/MM/YYYY", "DD.MM.YYYY", "DD/MM"]
    if format in allowed:
        try:
            if format=="YYYY-MM-DD": #YYYY-MM-DD
                d=iso.split("-")
                return date(int(d[0]), int(d[1]),  int(d[2]))
            if format=="DD/MM/YYYY": #DD/MM/YYYY
                d=iso.split("/")
                return date(int(d[2]), int(d[1]),  int(d[0]))
            if format=="DD.MM.YYYY": #DD.MM.YYYY
                d=iso.split(".")
                return date(int(d[2]), int(d[1]),  int(d[0]))
            if format=="DD/MM": #DD/MM
                d=iso.split("/")
                return date(date.today().year, int(d[1]),  int(d[0]))
        except:
            return None
    else:
        raise exceptions.CastException(_("I can't convert this format '{}'. I only support this {}").format(format, allowed))

def str2dtnaive(s, format):
    allowed=["%Y%m%d%H%M","%Y-%m-%d %H:%M:%S","%d/%m/%Y %H:%M","%d %m %H:%M %Y","%Y-%m-%d %H:%M:%S.","%H:%M:%S", '%b %d %H:%M:%S', "JsIso"]
    if format in allowed:
        if format=="%Y%m%d%H%M":
            dat=datetime.strptime( s, format )
            return dat
        if format=="%Y-%m-%d %H:%M:%S":#2017-11-20 23:00:00
            return datetime.strptime( s, format )
        if format=="%d/%m/%Y %H:%M":#20/11/2017 23:00
            return datetime.strptime( s, format )
        if format=="%d %m %H:%M %Y":#27 1 16:54 2017. 1 es el mes convertido con month2int
            return datetime.strptime( s, format)
        if format=="%Y-%m-%d %H:%M:%S.":#2017-11-20 23:00:00.000000  ==>  microsecond. Notice the point in format
            arrPunto=s.split(".")
            s=arrPunto[0]
            micro=int(arrPunto[1]) if len(arrPunto)==2 else 0
            dt=datetime.strptime( s, "%Y-%m-%d %H:%M:%S" )
            dt=dt+timedelta(microseconds=micro)
            return dt
        if format=="%H:%M:%S": 
            tod=date.today()
            a=s.split(":")
            return datetime(tod.year, tod.month, tod.day, int(a[0]), int(a[1]), int(a[2]))
        if format=='%b %d %H:%M:%S': #Apr 26 07:50:44. Year is missing so I set to current
            s=f"{date.today().year} {s}"
            return datetime.strptime(s, '%Y %b %d %H:%M:%S')
        if format=="JsIso": #2021-08-21T06:27:38.294
            s=s.replace("T"," ")
            dtnaive=str2dtnaive(s,"%Y-%m-%d %H:%M:%S.")
            return dtnaive
    else:
        raise exceptions.CastException(_("I can't convert this format '{}'. I only support this {}").format(format, allowed))

def str2dtaware(s, format, tz_name='UTC'):
    allowed=["%Y-%m-%d %H:%M:%S%z","%Y-%m-%d %H:%M:%S.%z", "JsUtcIso"]
    if format in allowed:
        if format=="%Y-%m-%d %H:%M:%S%z":#2017-11-20 23:00:00+00:00
            s=s[:-3]+s[-2:]
            dt=datetime.strptime( s, format )
            return dtaware_changes_tz(dt, tz_name)
        if format=="%Y-%m-%d %H:%M:%S.%z":#2017-11-20 23:00:00.000000+00:00  ==>  microsecond. Notice the point in format
            s=s[:-3]+s[-2:]#quita el :
            arrPunto=s.split(".")
            s=arrPunto[0]+s[-5:]
            micro=int(arrPunto[1][:-5])
            dt=datetime.strptime( s, "%Y-%m-%d %H:%M:%S%z" )
            dt=dt+timedelta(microseconds=micro)
            return dtaware_changes_tz(dt, tz_name)
        if format=="JsUtcIso": #2021-08-21T06:27:38.294Z
            s=s.replace("T"," ").replace("Z","")
            dtnaive=str2dtnaive(s,"%Y-%m-%d %H:%M:%S.")
            dtaware_utc=dtnaive2dtaware(dtnaive, 'UTC')
            return dtaware_changes_tz(dtaware_utc, tz_name)

#    else:
#        return timezone(tz_name).localize(str2dtnaive(s,format))

## epoch is the time from 1,1,1970 in UTC
## return now(timezone(self.name))
def dtaware2epochms(d):
    return d.timestamp()*1000
    
## Return a UTC datetime aware
def epochms2dtaware(n, tz="UTC"):
    utc_unaware=datetime.utcfromtimestamp(n/1000)
    utc_aware=utc_unaware.replace(tzinfo=ZoneInfo('UTC'))#Due to epoch is in UTC
    return dtaware_changes_tz(utc_aware, tz)

## epoch is the time from 1,1,1970 in UTC
## return now(timezone(self.name))
def dtaware2epochmicros(d):
    return int(d.timestamp()*1000000)
## Return a UTC datetime aware
def epochmicros2dtaware(n, tz="UTC"):
    utc_unaware=datetime.utcfromtimestamp(n/1000000)
    utc_aware=utc_unaware.replace(tzinfo=ZoneInfo('UTC'))#Due to epoch is in UTC
    return dtaware_changes_tz(utc_aware, tz)


## Returns a formated string of a dtaware string formatting with a zone name
## @param dt datetime aware object
## @return String
def dtaware2str(dt, format):
    if is_naive(dt)==True:
        raise exceptions.CastException("A datetime with timezone is needed for {}").format(dt)
    else:
        return dtnaive2str(dt, format)

## Returns a formated string of a dtaware string formatting with a zone name
## @param dt datetime aware object
## @param format String in ["%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y%m%d %H%M", "%Y%m%d%H%M"]
## @return String
def dtnaive2str(dt, format):
    allowed=["%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y%m%d %H%M", "%Y%m%d%H%M", "JsUtcIso"]
    if dt==None:
        return "None"
    elif format in allowed:
        if format=="%Y-%m-%d":
            return dt.strftime("%Y-%m-%d")
        elif format=="%Y-%m-%d %H:%M:%S": 
            return dt.strftime("%Y-%m-%d %H:%M:%S")
        elif format=="%Y%m%d %H%M": 
            return dt.strftime("%Y%m%d %H%M")
        elif format=="%Y%m%d%H%M":
            return dt.strftime("%Y%m%d%H%M")
        elif format=="JsUtcIso":
            return dt.strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        raise exceptions.CastException(_("I can't convert this format '{}'. I only support this {}").format(format, allowed))

## Changes zoneinfo from a dtaware object
## For example:
## - datetime.datetime(2018, 5, 18, 8, 12, tzinfo=<DstTzInfo 'Europe/Madrid' CEST+2:00:00 DST>)
## - libcaloriestrackerfunctions.dtaware_changes_tz(a,"Europe/London")
## - datetime.datetime(2018, 5, 18, 7, 12, tzinfo=<DstTzInfo 'Europe/London' BST+1:00:00 DST>)
## @param dt datetime aware object
## @param tzname String with datetime zone. For example: "Europe/Madrid"
## @return datetime aware object
def dtaware_changes_tz(dt,  tzname):
    if dt==None:
        return None
    if is_aware(dt):
        return dt.astimezone(ZoneInfo(tzname))
    else:
        raise exceptions.CastException(_("A datetime with timezone is needed"))

## Returns a list of tuples (year, month) from a month to another month, both included
## @param year_from Integer
## @param month_from Integer
## @param year_to Integer If none uses current year
## @param month_to Integer If none uses current month
def months(year_from, month_from, year_to=None, month_to=None):
    if year_to is None or month_to is None:
        year_to=date.today().year
        month_to=date.today().month
    r=[]
    end=date_first_of_the_month(year_to, month_to)
    current=date_first_of_the_month(year_from, month_from)
    while True:
        if current>end:
            break
        r.append((current.year,current.month))
        current=date_first_of_the_next_x_months(current.year, current.month, 1)
    return r



from collections import OrderedDict
from tabulate import tabulate
from gettext import translation
from importlib.resources import files
        
try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str

def lod_has_key(lod, key):
    if len(lod)==0:
        return False
    return key in lod[0]

## Order data columns. None values are set at the beginning
def lod_order_by(ld, key, reverse=False, none_at_top=True):
    nonull=[]
    null=[]
    for o in ld:
        com=o[key]
        if com is None:
            null.append(o)
        else:
            nonull.append(o)
    nonull=sorted(nonull, key=lambda item: item[key], reverse=reverse)
    if none_at_top==True:#Set None at top of the list
        return null+nonull
    else:
        return nonull+null

def lod_print(lod, number=None):
    """
    Function Prints a list of dictionaries with tabulate module.

    @param lod
    @type List of dictionaries
    @param number Number of dictionaries in the list to print. If None prints all lod. (defaults to None)
    @type Integer
    """
    number=len(lod) if number is None else number
    if len(lod)==0:
        print(_("This list of dictionaries hasn't data to print"))
    print(tabulate(lod[0:number], headers="keys", tablefmt="psql"))

def lod_sum(lod, key, ignore_nones=True):
    r=0
    for d in lod:
        if ignore_nones is True and d[key] is None:
            continue
        r=r+d[key]
    return r

def lod_sum_negatives(lod, key):
    r=0
    for d in lod:
        if d[key] is None or d[key]>0:
            continue
        r=r+d[key]
    return r

def lod_sum_positives(lod, key):
    r=0
    for d in lod:
        if d[key] is None or d[key]<0:
            continue
        r=r+d[key]
    return r

def lod_average(lod, key):
    return lod_sum(lod,key)/len(lod)

def lod_average_ponderated(lod, key_numbers, key_values):
    prods=0
    for d in lod:
        prods=prods+d[key_numbers]*d[key_values]
    return prods/lod_sum(lod, key_numbers)

def lod_median(lod, key):
    from statistics import median
    return median(lod2list(lod, key, sorted=True))

## Converts a lod to a dict using key as new dict key, and value as the key of the value field
def lod2dictkv(lod, key, value):
    d={}
    for ld in lod:
        d[ld[key]]=ld[value]
    return d

## Converts a lod to a dict using key as new dict key, and the dict as a value
def lod2dod(lod, key):
    d={}
    for ld in lod:
        d[ld[key]]=ld
    return d
    
## Converts a lod to a dict using (key1,key2) tuple  as new dict key, and the dict as a value
def lod2dod_tuple(lod, key1, key2):
    d={}
    for ld in lod:
        d[(ld[key1],ld[key2])]=ld
    return d

## Converts a dict of dictionaries (prefered orderdict) to a list of dictionaries
def dod2lod(d):
    r=[]
    for k,v in d.items():
        r.append(v)
    return r

## Returns a list from a lod key
## @param lod
## @param key String with the key to extract
## @param sorted Boolean. If true sorts final result
## @param cast String. "str", "float", casts the content of the key
def lod2list(lod, key, sorted=False, cast=None):
    r=[]
    for ld in lod:
        if cast is None:
            r.append(ld[key])
        elif cast == "str":
            r.append(str(ld[key]))
        elif cast == "float":
            r.append(float(ld[key]))
    if sorted is True:
        r.sort()
    return r

## Returns a list from a lod key, with distinct values, not all values
## @param lod
## @param key String with the key to extract
## @param sorted Boolean. If true sorts final result
## @param cast String. "str", "float", casts the content of the key
def lod2list_distinct(lod, key, sorted=False, cast=None):
    set_=set()
    for ld in lod:
        if cast is None:
            set_.add(ld[key])
        elif cast == "str":
            set_.add(str(ld[key]))
        elif cast == "float":
            set_.add(float(ld[key]))
    r=list(set_)
    if sorted is True:
        r.sort()
    return r


## Converts a list of ordereddict to a list of rows. ONLY DATA
## @params keys If None we must suppose is an ordered dict or keys will be randomized
def lod2lol(lod,  keys=None):
    if len(lod)==0:
        return []
        
    if keys is None:
        keys=lod[0].keys()
        
    r=[]  
    for od in lod:
        row_r=[]
        for key in keys:
            row_r.append(od[key])
        r.append(row_r)
    return r

def lod2lood(ld, keys):
    if len(ld)==0:
        return []
                
    r=[]  
    for d in ld:
        r_d=OrderedDict()
        for key in keys:
            r_d[key]=d[key]
        r.append(r_d)
    return r


## Returns maximum value of a given key. Is unique. REturns NOne if lod is empty
def lod_max_value(ld, key):
     if len(ld)>0:
          r=ld[0][key]
     else:
         return None
     for d in ld:
         if  d[key]>r:
             r=d[key]
     return r

## Returns minimum value of a given key. Is unique. REturns NOne if lod is empty
def lod_min_value(ld, key):
     if len(ld)>0:
          r=ld[0][key]
     else:
         return None
     for d in ld:
         if  d[key]<r:
             r=d[key]
     return r

def lod_rename_key(ld, from_, to_):
    """
        Renames a key(from_) to another key(to_) in each dictionary in ld
    """
    for d in ld:
        d[to_]=d.pop(from_)
    return ld

def lod_remove_key(lod, key):
    """
        Removes a key from all dictionaries in a list of dictionaries
    """
    for d in lod:
        del d[key]
    return lod
    
def lod_keys(lod_):
    """
        Return the keys of the list of dicts
        Return None if length of lod is 0. 
        Retuurn the keys of the first dictionary as a list
    """
    if len(lod_)>0:
        return list(lod_[0].keys())
    return None
    
def lod_clone(lod_):
    new_lod=[]
    for d in lod_:
        new_d={}
        for key, value in d.items():
            new_d[key]=value
        new_lod.append(new_d)
    return new_lod

def lod_filter_keys(lod_, keys):
    """
        Create a new lod leaving only keys in parameter
        @param keys List of keys to copy in new lod
    """
    new_lod=[]
    for d in lod_:
        new_d={}
        for key in keys:
            new_d[key]=d[key]
        new_lod.append(new_d)
    return new_lod


def lod_filter_dictionaries(lod_, lambda_function):
    """
        Create a new lod leaving filtering dictionaries that returns True to lambda funcion
        @param lambda_function needs two parameters (d, index) where d is the dictionary used to iterate and index is the position of the dictionary in the list of dictionaries
    """
    new_lod=[]
    for index,  d in enumerate(lod_):
        if lambda_function(d, index) is True:
            new_lod.append(d)
    return new_lod
    
def lod_calculate(lod_, key, lambda_function, clone=False):
    """
        Creates or replaces a column with the result of tthe lambda function in the lod passed as parameter
        @param key  Dictionary key where lambda_function result is going to be set
        @param lambda_function needs two parameters (d, index) where d is the dictionary used to iterate and index is the position of the dictionary in the list of dictionaries
        @param clone if True returns a clone of the lod_ parameter
    """
    if clone is True:
        l_o_d=lod_clone(lod_)
    else:
        l_o_d=lod_
        
    for index,  d in enumerate(l_o_d):
        d[key]=lambda_function(d, index)
    return l_o_d



    
    



from gettext import translation
from importlib.resources import files
from pydicts import exceptions
from tabulate import tabulate
        
try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str


## LOR is a list of list. Naned List Of Rows, used in myqtablewidget for example
## @param rows LOR
## @param index int with the index of the position where we are going to insert row
## @param column List with the values to add. Must be of the same size of rows
def lol_add_column(rows, index, column):
    if len(rows)!=len(column):
        raise exceptions.LolException(_("I can't add a column with different size of lol"))
    r_rows=[]
    for i, row in enumerate(rows):
        r_rows.append(row[0:index] + [column[i],] + row[index:len(row)])
    return r_rows

## Returns a list with object in positions removed
def list_remove_positions(l, listindex):
    if l is None:
        raise exceptions.LolException(_("I can't remove positions from a None list"))
    r=[]
    for i, o in enumerate(l):
        if i not in listindex:
            r.append(o)
    return r

## lol is a list of list. Naned List Of Rows, used in myqtablewidget
## @param listindex is a list of column indexes to remove
def lol_remove_columns(rows, listindex):
    r_rows=[]
    for i, row in enumerate(rows):
        r_rows.append(list_remove_positions(row,listindex))
    return r_rows

## lol is a list of list. Naned List Of Rows, used in myqtablewidget
## @param listindex is a list of column indexes to remove
def lol_remove_rows(rows, listindex):
    return list_remove_positions(rows, listindex) #It's a list but of row

## Return a lol transposed. Changed rows by columns
def lol_transposed(lol):
    if lol is None:
        raise exceptions.LolException(_("I can't traspaso a None object"))
    if len(lol)==0:
        return []
    r=[]
    columns=len(lol[0])
    for column in range(columns):
        tran_row=[]
        for row in lol:
            tran_row.append(row[column])
        r.append(tran_row)
    return r

## Extract a column from the list of row
def lol_get_row(lol, row):
    return lol[row]

## Extract a column from the list of row
def lol_get_column(lol, column):
    r=[]
    for row in lol:
        r.append(row[column])
    return r

## Return sum of values of a column from and index to and other index position. This method ignores None values
## This method can sum several objects
## @param row
## @param from_index
## @param to_index
## @param zerovalue 0 or Money(self.mem, 0, self.mem.localcurrency)....
def lol_sum_row(row, from_index, to_index, zerovalue=0):
    s=zerovalue
    for i, column in enumerate(row):
        if i>=from_index and i<=to_index:
            if column is not None:
                s=s + column
    return s

## Return sum of values of a column from and index to and other index position. This method ignores None values
## This method can sum several objects
## @zerovalue 0 or Money(self.mem, 0, self.mem.localcurrency)....
def lol_sum_column(lol, column, from_index, to_index, zerovalue=0):
    s=zerovalue
    for i, row in enumerate(lol):
        if i>=from_index and i<=to_index:
            if row[column] is not None:
                s=s + row[column]
    return s

def lol_print(lod, number=None):
    """
    Function Prints a list of list with tabulate module.

    @param lol
    @type List of lists
    @param number Number of lists to print. If None prints all lod. (defaults to None)
    @type Integer
    """
    number=len(lod) if number is None else number

    if len(lod)==0:
        print(_("lol_print: This list of lists hasn't data to print"))
        return
    if number==0:
        print(_("lol_print: No data was printed due to you selected 0 rows"))
        return
    print(tabulate(lod[0:number], tablefmt="psql"))


from datetime import date, time, datetime
from decimal import Decimal
from pydicts import casts, exceptions
from pytest import raises
from zoneinfo import ZoneInfo

zonename_madrid="Europe/Madrid"
zoneinfo_utc=ZoneInfo("UTC")
dtnaive=casts.dtnaive_now()
dtaware_utc=casts.dtaware_now()
dtaware_madrid=casts.dtaware_now(zonename_madrid)

def test_object_or_empty():
    assert casts.object_or_empty(None)==""
    assert casts.object_or_empty(1)==1
    
def test_str2decimal():
    assert casts.str2decimal(None, 1)==None
    assert casts.str2decimal("2.123,25", 1)==Decimal("2123.25")

def test_str2bool():
    with raises(exceptions.CastException):
        casts.str2bool(None)==None
    with raises(exceptions.CastException):
        assert casts.str2bool(1)==True
    with raises(exceptions.CastException):
        assert casts.str2bool(0)==False
    assert casts.str2bool("true")==True
    assert casts.str2bool("True")==True
    assert casts.str2bool("false")==False
    assert casts.str2bool("False")==False
    

def test_date_first_of_the_next_x_months():
    date_plus_3_months=casts.date_first_of_the_next_x_months(2023, 11, 3)
    assert date_plus_3_months.year==2024
    assert date_plus_3_months.month==2
    assert date_plus_3_months.day==1
    date_minus_12_months=casts.date_first_of_the_next_x_months(2023, 11, -12)
    assert date_minus_12_months.year==2022
    assert date_minus_12_months.month==11
    assert date_minus_12_months.day==1
    
def test_date_last_of_the_next_x_months():
    date_plus_3_months=casts.date_last_of_the_next_x_months(2023, 11, 3)
    assert date_plus_3_months.year==2024
    assert date_plus_3_months.month==2
    assert date_plus_3_months.day==29
    date_minus_12_months=casts.date_last_of_the_next_x_months(2023, 11, -12)
    assert date_minus_12_months.year==2022
    assert date_minus_12_months.month==11
    assert date_minus_12_months.day==30

def test_date_first_of_the_year():
    assert casts.date_first_of_the_year(2022)==date(2022, 1, 1)
    
def test_date_last_of_the_year():
    assert casts.date_last_of_the_year(2023)==date(2023, 12, 31)
    
### Function that converts a None value into a Decimal('0')
### @param dec Should be a Decimal value or None
### @return Decimal
#def test_none2decimal0(dec):
#    return none2alt(dec,Decimal('0'))
#
### If a value is None, returns an alternative
#def test_none2alt(value, alternative):
#    if value==None:
#        return alternative
#    return value
#

def test_bytes2str():
    assert casts.bytes2str(None)==None
    assert casts.bytes2str(b"Hello")=="Hello"
    
def test_str2bytes():
    assert casts.str2bytes(None)==None
    assert casts.str2bytes("Hello")==b"Hello"

def test_is_aware():
    assert casts.is_aware(dtnaive)==False
    assert casts.is_aware(dtaware_utc)==True

def test_is_naive():
    assert casts.is_naive(dtnaive)==True
    assert casts.is_naive(dtaware_utc)==False

def test_dtaware():
    assert dtaware_utc==casts.dtaware(dtaware_utc.date(), dtaware_utc.time(), "UTC")

def test_dtnaive2dtaware():
    naive=datetime(2023, 1, 1, 0, 0, 0, 0)
    assert casts.dtnaive2dtaware(naive, "Europe/Madrid")==naive.replace(tzinfo=ZoneInfo("Europe/Madrid"))

def test_date_first_of_the_month():
    assert casts.date_first_of_the_month(2023, 11)==date(2023, 11, 1)
    with raises(ValueError):
        casts.date_first_of_the_month(2023, 13)

def test_date_last_of_the_month():
    assert casts.date_last_of_the_month(2023, 11)==date(2023, 11, 30)
    with raises(ValueError):
        casts.date_first_of_the_month(2023, 13)

def test_dtaware_month_end():
    assert casts.dtaware_month_end(2023, 11, "UTC")==datetime(2023, 11, 30, 23, 59, 59, 999999, ZoneInfo("UTC"))
    
def test_dtaware_year_start():
    assert casts.dtaware_year_start(2023, "UTC")==datetime(2023, 1, 1,0, 0, 0, 0, ZoneInfo("UTC"))

def test_dtaware_year_end():
    assert casts.dtaware_year_end(2023, "UTC")==datetime(2023, 12, 31, 23, 59, 59, 999999, ZoneInfo("UTC"))

def test_dtnaive_day_end():
    assert casts.dtnaive_day_end(casts.dtnaive_now()).time()==time(23, 59, 59, 999999)

def test_dtnaive_day_end_from_date():
    assert casts.dtnaive_day_end_from_date(date(2023, 12, 29))==datetime(2023, 12, 29, 23, 59, 59, 999999)

def test_dtaware_day_end_from_date():
    assert casts.dtaware_day_end_from_date(date(2023, 12, 29), "UTC")==datetime(2023, 12, 29, 23, 59, 59, 999999, ZoneInfo("UTC"))

### Returns a dtnaive or dtawre (as parameter) with the end of the day
#def test_dt_day_start(dt):
#    return dt.replace(hour=0, minute=0, second=0, microsecond=0)
#
### Returns the end of the day dtnaive from a date
#def test_dtnaive_day_start_from_date(dat):
#    dt=datetime(dat.year, dat.month, dat.day)
#    return dt_day_start(dt)
#
### Returns the end of the day dtaware of the tz_name timezone from a date
#def test_dtaware_day_start_from_date(date, tz_name):
#    dt=dtaware(date, time(0, 0), tz_name)
#    return dt_day_start(dt)
#
### Returns the start of a month
#def test_dtaware_month_start(year, month, tz_name):
#    return dtaware_day_start_from_date(date(year, month, 1), tz_name)
#    
#def test_month2int(s):
#    """
#        Converts a month string to a int
#    """
#    if s in ["Jan", "Ene", "Enero", "January", "enero", "january"]:
#        return 1
#    if s in ["Feb", "Febrero", "February", "febrero", "february"]:
#        return 2
#    if s in ["Mar", "Marzo", "March", "marzo", "march"]:
#        return 3
#    if s in ["Apr", "Abr", "April", "Abril", "abril", "april"]:
#        return 4
#    if s in ["May", "Mayo", "mayo", "may"]:
#        return 5
#    if s in ["Jun", "June", "Junio", "junio", "june"]:
#        return 6
#    if s in ["Jul", "July", "Julio", "julio", "july"]:
#        return 7
#    if s in ["Aug", "Ago", "August", "Agosto", "agosto", "august"]:
#        return 8
#    if s in ["Sep", "Septiembre", "September", "septiembre", "september"]:
#        return 9
#    if s in ["Oct", "October", "Octubre", "octubre", "october"]:
#        return 10
#    if s in ["Nov", "Noviembre", "November", "noviembre", "november"]:
#        return 11
#    if s in ["Dic", "Dec", "Diciembre", "December", "diciembre", "december"]:
#        return 12
#
def test_str2time():
#    allowed=["HH:MM", "HH:MM:SS","HH:MMxx"]
    assert casts.str2time("09:05", "HH:MM")==time(9, 5)
    assert casts.str2time("09:05:54", "HH:MM:SS")==time(9, 5, 54)
    assert casts.str2time("09:05pm", "HH:MMxx")==time(21, 5)

def test_time2str():
#    allowed=["HH:MM", "HH:MM:SS","Xulpymoney"]
    time_=time(9, 5, 54)
    assert casts.time2str(time_, "HH:MM")=="09:05"
    assert casts.time2str(time_, "HH:MM:SS")=="09:05:54"

def test_str2date():
#    allowed=["YYYY-MM-DD", "DD/MM/YYYY", "DD.MM.YYYY", "DD/MM"]

    assert casts.str2date("2023-11-26")==date(2023, 11, 26)
    assert casts.str2date("2023-11-26", "YYYY-MM-DD")==date(2023, 11, 26)
    assert casts.str2date("26/11/2023", "DD/MM/YYYY")==date(2023, 11, 26)
    assert casts.str2date("26.11.2023", "DD.MM.YYYY")==date(2023, 11, 26)
    assert casts.str2date("26/11", "DD/MM")==date(2023, 11, 26)
#
def test_str2dtnaive():
#    allowed=["%Y%m%d%H%M","%Y-%m-%d %H:%M:%S","%d/%m/%Y %H:%M","%d %m %H:%M %Y","%Y-%m-%d %H:%M:%S.","%H:%M:%S", '%b %d %H:%M:%S']

#    assert casts.str2dtnaive("2023-11-26", "%Y-%m-%d")==datetime(2023, 11, 26)
#    assert casts.str2dtnaive("2023-11-26 17:05:05", "%Y-%m-%d %H:%M:%S")==datetime(2023, 11, 26, 17, 5, 5)
#    assert casts.str2dtnaive("20231126 1705", "%Y%m%d %H%M")==datetime(2023, 11, 26, 17, 5, 5)
    assert casts.str2dtnaive("202311261705", "%Y%m%d%H%M")==datetime(2023, 11, 26, 17, 5)
    assert casts.str2dtnaive("2023-11-26T17:05:05.123456", "JsIso")==datetime(2023, 11, 26, 17, 5, 5, 123456)
    assert casts.str2dtnaive("2023-11-26T17:05:05", "JsIso")==datetime(2023, 11, 26, 17, 5, 5)


def test_str2dtaware():
#    allowed=["%Y-%m-%d %H:%M:%S%z","%Y-%m-%d %H:%M:%S.%z", "JsUtcIso"]
    #assert casts.str2dtaware("2023-11-26", "%Y-%m-%d")==datetime(2023, 11, 26, tzinfo=ZoneInfo('UTC'))
    #assert casts.str2dtaware("202311261705", "%Y%m%d%H%M")==datetime(2023, 11, 26, 17, 5)
    assert casts.str2dtaware("2023-11-26T17:05:05Z", "JsUtcIso")==datetime(2023, 11, 26, 17, 5, 5, tzinfo=ZoneInfo('UTC'))

### epoch is the time from 1,1,1970 in UTC
### return now(timezone(self.name))
#def test_dtaware2epochms(d):
#    return d.timestamp()*1000
#    
### Return a UTC datetime aware
#def test_epochms2dtaware(n, tz="UTC"):
#    utc_unaware=datetime.utcfromtimestamp(n/1000)
#    utc_aware=utc_unaware.replace(tzinfo=timezone('UTC'))#Due to epoch is in UTC
#    return dtaware_changes_tz(utc_aware, tz)
#
### epoch is the time from 1,1,1970 in UTC
### return now(timezone(self.name))
#def test_dtaware2epochmicros(d):
#    return int(d.timestamp()*1000000)
### Return a UTC datetime aware
#def test_epochmicros2dtaware(n, tz="UTC"):
#    utc_unaware=datetime.utcfromtimestamp(n/1000000)
#    utc_aware=utc_unaware.replace(tzinfo=timezone('UTC'))#Due to epoch is in UTC
#    return dtaware_changes_tz(utc_aware, tz)
#
#

def test_dtaware2string():    
    dt_aware=datetime(2023, 11, 26, 17, 5, 5, 123456, tzinfo=ZoneInfo("Europe/Madrid"))
    assert casts.dtnaive2str(dt_aware, "%Y-%m-%d")=="2023-11-26"
    assert casts.dtnaive2str(dt_aware, "%Y-%m-%d %H:%M:%S")=="2023-11-26 17:05:05"
    assert casts.dtnaive2str(dt_aware, "%Y%m%d %H%M")=="20231126 1705"
    assert casts.dtnaive2str(dt_aware, "%Y%m%d%H%M")=="202311261705"
    assert casts.dtnaive2str(dt_aware, "JsUtcIso")=="2023-11-26T17:05:05Z"


def test_dtnaive2string():
    dt_naive=datetime(2023, 11, 26, 17, 5, 5, 123456)
    assert casts.dtnaive2str(dt_naive, "%Y-%m-%d")=="2023-11-26"
    assert casts.dtnaive2str(dt_naive, "%Y-%m-%d %H:%M:%S")=="2023-11-26 17:05:05"
    assert casts.dtnaive2str(dt_naive, "%Y%m%d %H%M")=="20231126 1705"
    assert casts.dtnaive2str(dt_naive, "%Y%m%d%H%M")=="202311261705"
    assert casts.dtnaive2str(dt_naive, "JsUtcIso")=="2023-11-26T17:05:05Z"

def test_dtaware_changes_tz():
    #Sacado date en linux
    dt_madrid=datetime(2023, 11, 26, 12, 0,  tzinfo=ZoneInfo("Europe/Madrid"))
    dt_utc=datetime(2023, 11, 26, 11, 0,  tzinfo=ZoneInfo("UTC"))
    
    assert casts.dtaware_changes_tz(dt_madrid, "UTC")==dt_utc
    assert casts.dtaware_changes_tz(dt_utc, "Europe/Madrid")==dt_madrid

def test_months():
    assert casts.months(2023, 11, 2024, 1)== [(2023, 11), (2023, 12), (2024, 1)]
    assert casts.months(2023, 9,  2023, 9)== [(2023, 9)]




from decimal import Decimal
from datetime import date, datetime, time, timedelta
from zoneinfo import ZoneInfo
from pydicts import myjsonencoder
d={}
d["None"]=None
d["Integer"]=12112
d["Float"]=12121.1212
d["Date"]=date.today()
d["Datetime"]=datetime.now()
d["Timedelta"]=timedelta(hours=4, days=2, minutes=12,  seconds=12)
d["Time"]=time(12, 12, 12, 123456)
d["Datetime aware"]=d["Datetime"].replace(tzinfo=ZoneInfo("UTC"))
d["Bytes"]=b"Byte array"
d["Decimal"]=Decimal("12.12123414")

def test_myjsonencoder():
    json_string=myjsonencoder.MyJSONEncoder_dumps(d)
    json_=myjsonencoder.MyJSONEncoder_loads(json_string)
    assert json_["None"]==d["None"]
    assert json_["Integer"]==d["Integer"]
    assert json_["Float"]==d["Float"]
    assert json_["Date"]==d["Date"]
    assert json_["Datetime"]==d["Datetime"]
    assert json_["Datetime aware"]==d["Datetime aware"]
    assert json_["Bytes"]==d["Bytes"]
    assert json_["Decimal"]==d["Decimal"]
#    assert json_["Time"]==d["Time"]
#    assert json_["Timedelta"]==d["Timedelta"] 
    
def test_myjsonencoder_decimals_as_float():
    json_string=myjsonencoder.MyJSONEncoderDecimalsAsFloat_dumps(d)
    json_=myjsonencoder.MyJSONEncoderDecimalsAsFloat_loads(json_string)
    assert json_["None"]==d["None"]
    assert json_["Integer"]==d["Integer"]
    assert json_["Float"]==d["Float"]
    assert json_["Date"]==d["Date"]
    assert json_["Datetime"]==d["Datetime"]
    assert json_["Datetime aware"]==d["Datetime aware"]
    assert json_["Bytes"]==d["Bytes"]
    assert json_["Decimal"]==float(d["Decimal"])
#    assert json_["Time"]==d["Time"]
#    assert json_["Timedelta"]==d["Timedelta"] 


def test_myjsonencoder_decimals_as_string():
    json_string=myjsonencoder.MyJSONEncoderDecimalsAsString_dumps(d)
    json_=myjsonencoder.MyJSONEncoderDecimalsAsString_loads(json_string)
    assert json_["None"]==d["None"]
    assert json_["Integer"]==d["Integer"]
    assert json_["Float"]==d["Float"]
    assert json_["Date"]==d["Date"]
    assert json_["Datetime"]==d["Datetime"]
    assert json_["Datetime aware"]==d["Datetime aware"]
    assert json_["Bytes"]==d["Bytes"]
    assert json_["Decimal"]==str(d["Decimal"])
#    assert json_["Time"]==d["Time"]
#    assert json_["Timedelta"]==d["Timedelta"] 



from pydicts.lod import lod_has_key, lod_print, lod_print_first, lod_sum, lod2list, lod_average_ponderated
from gettext import translation
from importlib.resources import files
        
try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str
## El objetivo es crear un objeto list_dict que se almacenera en self.ld con funciones set
## set_from_db #Todo se carga desde base de datos con el minimo parametro posible
## set_from_db_and_variables #Preguntara a base datos aquellas variables que falten. Aunque no estén en los parámetros p.e. money_convert
## set_from_variables #Solo con variables
## set #El lod ya está hecho pero se necesita el objeto para operar con el
##class Do:
##    def __init__(self,d):
##        self.d=d
##        self.create_attributes()
##
##    def number_keys(self):
##        return len(self.d)
##
##    def has_key(self,key):
##        return key in self.d
##
##    def print(self):
##        lod_print(self.d)
##
##    ## Creates an attibute from a key
##    def create_attributes(self):
##        for key, value in self.d.items():
##            setattr(self, key, value)




## Class that return a object to manage lod
## El objetivo es crear un objeto list_dict que se almacenera en self.ld con funciones set
## set_from_db #Todo se carga desde base de datos con el minimo parametro posible
## set_from_db_and_variables #Preguntara a base datos aquellas variables que falten. Aunque no estén en los parámetros p.e. money_convert
## set_from_variables #Solo con variables
## set #El lod ya está hecho pero se necesita el objeto para operar con el
class LOD:
    def __init__(self, name=None):
        self.name=self.__class__.__name__ if name is None else name
        self.ld=[]

    def length(self):
        return len(self.ld)

    def has_key(self,key):
        return lod_has_key(self.ld,key)

    def print(self):
        lod_print(self.ld)

    def print_first(self):
        lod_print_first(self.ld)

    def sum(self, key, ignore_nones=True):
        return lod_sum(self.ld, key, ignore_nones)

    def list(self, key, sorted=True):
        return lod2list(self.ld, key, sorted)

    def average_ponderated(self, key_numbers, key_values):
        return lod_average_ponderated(self.ld, key_numbers, key_values)

    def set(self, ld):
        del self.ld
        self.ld=ld
        return self

    def is_set(self):
        if hasattr(self, "ld"):
            return True
        print(_("You must set your lod in {}").format(self.name))
        return False

    def append(self,o):
        self.ld.append(o)

    def first(self):
        return self.ld[0] if self.length()>0 else None

    ## Return list keys of the first element[21~
    def first_keys(self):
        if self.length()>0:
            return self.first().keys()
        else:
            return "I can't show keys"
    
    def order_by(self, key, reverse=False):
        self.ld=sorted(self.ld,  key=lambda item: item[key], reverse=reverse)



#from datetime import datetime, date
#from decimal import Decimal
from pydicts import lol, exceptions
from pytest import raises


lol_=[]
for i in range(10):
    lol_.append([1*i,2*i,3*i])

def test_lol_add_column():
    lol.lol_add_column(lol_, 2,  [1, 2, 3, 4, 5, 6, 7, 8, 9, 0])

def test_lol_print():
    lol.lol_print(lol_)
    lol.lol_print(lol_,0)
    lol.lol_print(lol_,-1)
    
def test_lol_transposed():
    transposed=lol.lol_transposed(lol_)
    transposed=lol.lol_transposed([])
    with raises(exceptions.LolException):
        transposed=lol.lol_transposed(None)

#
#    lor=[]
#    column_to_add=[]
#    for i in range(10):
#        lor.append([1*i,2*i,3*i])
#        column_to_add.append(-i)
#    print_lor(lor)
#
#    lor=lor_add_column(lor, 0, column_to_add)
#    lor=lor_add_column(lor, 2, column_to_add)
#    lor=lor_add_column(lor, 5, column_to_add)
#    print_lor(lor)
#
#    a=lor_remove_columns(lor,[2,3])
#    print_lor(a)
#    b=lor_remove_rows(a,[8,9])
#    print_lor(b)
#
#    c=lor_transposed(b)
#    print_lor(c)
#
#    d=Decimal("12.3")
#    json_d=var2json(d)
#    print (d, json_d, json_d.__class__)
#    d=None
#    json_d=var2json(d)
#    print (d, json_d, json_d.__class__)
#    d=True
#    json_d=var2json(d)
#    print (d, json_d, json_d.__class__)



from pydicts import __version__
from os import system

def release():
    print("""Nueva versión:
  * Cambiar la version en pyproject.toml
  * Cambiar la versión y la fecha en __init__.py
  * Modificar el Changelog en README
  * poe translate
  * linguist
  * poe translate
  * poe test
  * git commit -a -m 'pydicts-{0}'
  * git push
  * Hacer un nuevo tag en GitHub
  * poetry build
  * poetry publish
  * Crea un nuevo ebuild de pydicts en Gentoo con la nueva versión
  * Subelo al repositorio myportage

""".format(__version__))


def translate():
    #es
    system("xgettext -L Python --no-wrap --no-location --from-code='UTF-8' -o pydicts/locale/pydicts.pot pydicts/*.py")
    system("msgmerge -N --no-wrap -U pydicts/locale/es.po pydicts/locale/pydicts.pot")
    system("msgfmt -cv -o pydicts/locale/es/LC_MESSAGES/pydicts.mo pydicts/locale/es.po")
    system("msgfmt -cv -o pydicts/locale/en/LC_MESSAGES/pydicts.mo pydicts/locale/en.po")




from shutil import which

if which("pdflatex") is not None:
    from pydicts import pylatex
    from pylatex import Document, Section, Subsection
    from pylatex.package import Package
    from pylatex.utils import italic, NoEscape
    lod=[
            {"year": 2022, "month": 1, "my_sum": 12},
            {"year": 2021, "month": 2, "my_sum": 123},
            {"year": 2019, "month": 5, "my_sum": 1},
            {"year": 2022, "month": 12, "my_sum": 12},
    ]
    def tests_pylatex_table_header():

        doc = Document()
        doc.packages.append(Package('xcolor'))

        with doc.create(Section('A section')):
            doc.append('Some regular text and some ')
            doc.append(italic('italic text. '))

            with doc.create(Subsection('A subsection')):
                doc.append('Also some crazy characters: $&#{}')
                
        doc.append(NoEscape("\\centering"))
        pylatex.pylatex_table(doc, lod)
        pylatex.pylatex_table(doc, [])
        pylatex.pylatex_table_with_matched_values(doc,  [2022, 2, 12], lod, code_="|l|c|r|", match_color="teal", unmatch_color="red")
        doc.generate_pdf('test_pylatex_table_header', clean_tex=False)
